{------------------------------------------------------------------------------
   _____  ____  ____  ____
  (  _  )(_  _)(_  _)( ___) Object-oriented
   )(_)(   )(    )(   )__)  Tengwar
  (_____) (__)  (__) (____) Transcription
  version 0.1.1             Engine

  by Michal Nowakowski <miszka@limes.com.pl>

  OTTE is an ObjectPascal unit that provides you with an easy way to
  create a transcriber (not translator!) from Latin alphabet to Tolkien's
  Tengwar and back, like famous TengScribe or YaTT, in Delphi or FPC.

  Although OTTE's transcription back from Tengwar to Latin could be much
  better for modes that place tehtar over the following consonant (like
  English or Sindarin Classic), and does not support Narmacil's TMF modes,
  OTTE is much faster even than Narmacil and still in development,
  so there may be new things added and old improved.

  I have not made any documentation yet... For now, see the program TENG.
  I also plan to make a shared linux library with C and pascal headers so
  that others can use OTTE too.

                                --*--

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

------------------------------------------------------------------------------}

{$I otte_cfg.pas}
unit otte;

interface //===================================================================

uses
  otte_math,
  otte_ctype;

const
  CotteOK              =  0;
  CotteErrSetCmapFailed= -1;
  CotteErrModNotOpened = -2;
  CotteErrModReadFault = -3;
  CotteErrModTooLarge  = -4;
  CotteErrModBadFormat = -5;
  CotteErrNotEnoughRAM = -6;
  CotteErrModNotLoaded = -7;

type
  TotteResult = CotteErrModNotLoaded .. CotteOK;

const
  CottePrimaryAlphabet = true;
  CotteSecondaryAlphabet = false;

  CotteRoman = CottePrimaryAlphabet;
  CotteTengwar = CotteSecondaryAlphabet;

type
  TotteAlphabet = boolean;

const
  CotteTengwarQuestionMark = #174;
  CotteRomanQuestionMark = '?';
  CotteNoQuestionMark = #00;

  CotteRomanDigits = '0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ';
  CotteRomanNumBase = 10;
  CotteTengwarDigits = #240#241#242#243#244#245#246#247#248#249#250#251;
  CotteTengwarNumBase = 12;

  CotteModStringLen = 14;
  CotteModClauseSize = 36;
  CotteModDescLen = CotteModClauseSize-2;
  CotteModMaxClauses = 1500;

  CotteAnyChar = #00;
  CotteNoAlpha = #13;

type
  TotteCharClass = class(TObject)
    public
      IsAlpha, IsDigit, IsNonSpacing: array[TotteAlphabet] of TotteCharProperty;
      DigitValues: array[TotteAlphabet, char] of byte;
      Digits: array[TotteAlphabet] of string[255];
      NumBase: array[TotteAlphabet] of byte;
      AdjustChar: array[TotteAlphabet] of TotteCharAdjust;
      RomanQuestionMark, TengwarQuestionMark: char;
      constructor create;
      procedure reset;
      procedure SetDigits(Alphabet: TotteAlphabet; const s: string; NumBase: byte);
      function SetCharmap(const charmap: ansistring): integer;
      function AdjustStr(Alphabet: TotteAlphabet; const s: ansistring): ansistring;
  end;

  TotteEncodedModClause = array[0..CotteModClauseSize-1] of byte;
  TotteEncodedModClauses = array of TotteEncodedModClause;

  TotteModDesc = string[CotteModDescLen];
  TotteModString = string[CotteModStringLen];

  TotteModClause = record
    Roman, Tengwar: TotteModString;
    NextRoman, LastTengwa, SetLastTengwa: char;
  end;

  TotteModClauses = array of TotteModClause;
  TotteModSortControl = function(const a, b: TotteModClause): boolean;

  TotteMode = class(TotteCharClass)
    protected
      RawMode: TotteEncodedModClauses;
      procedure SortClauses(var clauses: TotteModClauses; const control: TotteModSortControl);
    public
      Description: TotteModDesc;
      CaseSensitive, TengwaCaseSensitive: boolean;
      Clauses, BackClauses: TotteModClauses;
      constructor create;
      destructor destroy; override;
      procedure reset;
      procedure sort(Alphabet: TotteAlphabet; const control: TotteModSortControl);
      function LoadFromFile(const fname: ansistring): integer;
    //function LoadFromStream(stream: TStream): integer;
      function reload: integer;
    //function SaveToFile(const fname: ansistring): integer;
    //function SaveToStream(stream: TStream): integer;
      procedure AutoBackClauses;
  end;

  TotteCharIndex = array[char] of integer;

  TotteTranscriber = class(TotteMode)
    protected
      FirstInd, LastInd, FirstBackInd, LastBackInd: TotteCharIndex;
      IsWhiteSpace: TotteCharProperty;
      DoTranscription: boolean;
    public
      RevNum: boolean;        //reverse digit order
      CirNum: boolean;        //mark duedecades with circles beneath
      DotNum: boolean;        //mark numerals with dots
      BarNum: boolean;        //mark numerals with bars
      B12Num: boolean;        //use duodecimal numeric base
      SepNum: boolean;        //treat numbers or spaces separately
      procedure reset;
      constructor create;
      function GetReady: integer;  //run this after any clauses modification
      function LoadFromFile(const fname: ansistring): integer;
    //function LoadFromStream(stream: TStream): integer; override;
      function TranscribeLine(TranscribeBack: boolean; const s: ansistring): ansistring;
      procedure StopTranscription;
  end;

const
  CotteIsTengwar: TotteCharProperty =
  ( false, false, false, false, false, false, false, false,
    false, false, false, false, false, false, false, false,
    false, false, false, false, false, false, false, false,
    false, false, false, false, false, false, false, false,
    false, true,  true,  true,  true,  true,  true,  true, 
    true,  true,  true,  true,  true,  false, true,  true, 
    true,  true,  true,  true,  true,  true,  true,  true, 
    true,  true,  true,  true,  true,  false, true,  true, 
    true,  true,  true,  true,  true,  true,  true,  true, 
    true,  true,  true,  true,  true,  true,  true,  true, 
    true,  true,  true,  true,  true,  true,  true,  true, 
    true,  true,  true,  true,  true,  true,  true,  true, 
    true,  true,  true,  true,  true,  true,  true,  true, 
    true,  true,  true,  true,  true,  true,  true,  true, 
    true,  true,  true,  true,  true,  true,  true,  true, 
    true,  true,  true,  true,  true,  true,  true,  false,
    false, false, true,  true,  true,  true,  true,  true, 
    false, true,  true,  true,  false, false, false, false,
    false, true,  true,  true,  true,  true,  true,  true, 
    true,  true,  true,  false, false, false, false, true, 
    false, true,  true,  true,  false, true,  true,  true, 
    true,  true,  true,  false, false, true,  false, true, 
    true,  false, false, true,  true,  true,  false, false,
    true,  false, true,  false, true,  true,  true,  true, 
    false, false, false, false, true,  true,  true,  false,
    true,  true,  true,  true,  true,  true,  true,  true, 
    true,  true,  true,  true,  true,  true,  true,  true, 
    true,  true,  true,  true,  true,  true,  true,  true, 
    true,  true,  true,  true,  true,  true,  true,  true, 
    true,  true,  true,  true,  true,  true,  true,  true, 
    false, false, false, false, false, false, false, false,
    false, false, false, false, true,  true,  true,  true  );

  CotteIsNonSpacingTehta: TotteCharProperty =
  ( false, false, false, false, false, false, false, false,
    false, false, false, false, false, false, false, false,
    false, false, false, false, false, false, false, false,
    false, false, false, false, false, false, false, false,
    false, false, false, true,  true,  true,  true,  true,
    true,  true,  false, false, false, false, false, true,
    true,  false, false, false, false, false, false, false,
    false, false, true,  true,  false, false, true,  true,
    false, false, true,  true,  true,  true,  true,  true,
    true,  false, true,  false, true,  true,  true,  true,
    true,  false, true,  false, true,  true,  true,  false,
    false, true,  false, true,  false, false, true,  false,
    false, false, false, false, false, false, false, false,
    false, false, false, false, false, false, false, false,
    true,  false, false, false, false, false, false, false,
    false, false, false, true,  true,  true,  false, false,
    false, false, true,  true,  true,  true,  true,  false,
    false, true,  true,  true,  false, false, false, false,
    false, true,  true,  true,  true,  false, false, false,
    true,  true,  false, false, false, false, false, true,
    true,  false, false, false, false, false, false, false,
    true,  true,  true,  false, false, true,  false, true,
    true,  false, false, false, true,  true,  false, false,
    true,  false, false, false, false, false, false, false,
    false, false, false, false, false, false, false, false,
    true,  true,  true,  true,  true,  true,  true,  true,
    true,  true,  true,  true,  true,  true,  true,  true,
    true,  true,  true,  true,  true,  true,  true,  true,
    true,  true,  true,  true,  true,  true,  true,  true,
    true,  true,  true,  true,  true,  true,  true,  true,
    false, false, false, false, false, false, false, false,
    false, false, false, false, true,  true,  true,  true   );

  CotteAdjustTehtar: TotteCharAdjust =
  ( #0,    #1,    #2,    #3,    #4,    #5,    #6,    #7,
    #8,    #9,    #10,   #11,   #12,   #13,   #14,   #15,
    #16,   #17,   #18,   #19,   #20,   #21,   #22,   #23,
    #24,   #25,   #26,   #27,   #28,   #29,   #30,   #31,
    #32,   #33,   #39,   #35,   #36,   #37,   #38,   #39,
    #40,   #48,   #42,   #56,   #44,   #45,   #46,   #59,
    #48,   #49,   #50,   #51,   #52,   #53,   #54,   #55,
    #56,   #57,   #59,   #59,   #60,   #61,   #62,   #59,
    #64,   #65,   #37,   #35,   #35,   #35,   #36,   #37,
    #94,   #56,   #38,   #75,   #76,   #38,   #94,   #79,
    #48,   #81,   #36,   #83,   #37,   #38,   #36,   #87,
    #88,   #94,   #90,   #91,   #92,   #93,   #94,   #56,
    #96,   #97,   #98,   #99,   #100,  #101,  #102,  #103,
    #104,  #105,  #106,  #107,  #108,  #109,  #110,  #111,
    #48,   #113,  #114,  #115,  #116,  #117,  #118,  #119,
    #120,  #121,  #122,  #91,   #56,   #56,   #126,  #127,
    #128,  #129,  #130,  #131,  #131,  #131,  #131,  #135,
    #136,  #137,  #138,  #139,  #140,  #141,  #142,  #143,
    #144,  #145,  #145,  #145,  #145,  #149,  #150,  #151,
    #152,  #153,  #154,  #155,  #156,  #157,  #158,  #159,
    #160,  #56,   #56,   #56,   #164,  #56,   #166,  #167,
    #168,  #169,  #170,  #171,  #172,  #170,  #174,  #170,
    #176,  #177,  #178,  #179,  #180,  #170,  #182,  #183,
    #184,  #185,  #186,  #187,  #188,  #189,  #190,  #191,
    #192,  #193,  #194,  #195,  #196,  #56,   #56,   #199,
    #200,  #201,  #202,  #203,  #204,  #205,  #206,  #207,
    #62,   #62,   #62,   #62,   #212,  #212,  #212,  #212,
    #216,  #216,  #216,  #216,  #220,  #220,  #220,  #220,
    #224,  #224,  #224,  #224,  #228,  #228,  #228,  #228,
    #232,  #232,  #232,  #232,  #91,   #39,   #91,   #39,
    #240,  #241,  #242,  #243,  #244,  #245,  #246,  #247,
    #248,  #249,  #250,  #251,  #252,  #252,  #252,  #252 );

implementation

//===TotteCharClass============================================================
constructor TotteCharClass.create;
begin
  inherited;
  self.reset;
end;

//-----------------------------------------------------------------------------
procedure TotteCharClass.reset;
begin
  SetCharmap('');
  fillchar(IsNonSpacing[CotteRoman], sizeof(TotteCharProperty), false);
  RomanQuestionMark := CotteRomanQuestionMark;
  SetDigits(CotteRoman, CotteRomanDigits, CotteRomanNumBase);

  IsAlpha[CotteTengwar] := CotteIsTengwar;
  AdjustChar[CotteTengwar] := CotteAdjustTehtar;
  IsNonSpacing[CotteTengwar] := CotteIsNonSpacingTehta;
  TengwarQuestionMark := CotteTengwarQuestionMark;
  SetDigits(CotteTengwar, CotteTengwarDigits, CotteTengwarNumBase);
end;

//-----------------------------------------------------------------------------
procedure TotteCharClass.SetDigits(Alphabet: TotteAlphabet; const s: string; NumBase: byte);

var
  i, base: byte;

begin
  fillchar(IsDigit[Alphabet], sizeof(TotteCharProperty), false);
  fillchar(DigitValues[Alphabet], sizeof(DigitValues[Alphabet]), 255);
  fillchar(Digits[Alphabet], sizeof(Digits[Alphabet]), ' ');

  if length(s) < NumBase then base := length(s)
  else base := NumBase;

  for i := 1 to base do IsDigit[Alphabet, s[i]] := true;
  for i := 1 to length(s) do DigitValues[Alphabet, s[i]] := i;

  self.NumBase[Alphabet] := base;
  Digits[Alphabet] := s;
end;

//-----------------------------------------------------------------------------
function TotteCharClass.SetCharmap(const charmap: ansistring): integer;
begin
  if otte_ctype.load_charmap(
    charmap, IsAlpha[CotteRoman], AdjustChar[CotteRoman])
  then result := CotteOK
  else result := CotteErrSetCmapFailed;
end;

//-----------------------------------------------------------------------------
function TotteCharClass.AdjustStr(Alphabet: TotteAlphabet; const s: ansistring): ansistring;

var
  i: integer;
  p: ^TotteCharAdjust;

begin
  p := @AdjustChar[Alphabet];
  result := s;
  for i := 1 to length(s) do result[i] := p[s[i]]
end;

//===TotteMode==================================================================
procedure TotteMode.reset;
begin
  inherited;
  SetLength(RawMode, 0);  
  Description := '';
  CaseSensitive := false;
  TengwaCaseSensitive := false;
  SetLength(clauses, 0);
  SetLength(BackClauses, 0);
end;

//-----------------------------------------------------------------------------
constructor TotteMode.create;
begin
  inherited;
  reset;
end;

//-----------------------------------------------------------------------------
destructor TotteMode.destroy;
begin
  reset;
  inherited;
end;

//-----------------------------------------------------------------------------
procedure TotteMode.SortClauses(var clauses: TotteModClauses; const control: TotteModSortControl);

var
  i, j: integer;
  t: TotteModClause;

{ Do NOT use any unstable sorting method!
  Also, Insertionsort is fast enough.
}

begin
  if Addr(control) = nil then exit
  else for i := 1 to high(clauses) do begin
    j := i;
    t := clauses[j];

    while (j>0) and control(clauses[j-1], t) do begin
      clauses[j] := clauses[j-1];
      dec(j);
    end;

    clauses[j] := t;
  end;
end;

//-----------------------------------------------------------------------------
procedure TotteMode.sort(Alphabet: TotteAlphabet; const control: TotteModSortControl);
begin
  if Alphabet then self.SortClauses(self.Clauses, control)
  else self.SortClauses(self.BackClauses, control);
end;

//-----------------------------------------------------------------------------
function TotteMode.LoadFromFile(const fname: ansistring): integer;

var
  f: file;
  fsize, count: integer;
  data, swap: TotteEncodedModClauses;
  ClausesCount: integer;

begin
  result := CotteOK;
  swap := nil;
  data := nil;

  assign(f, fname);
  system.reset(f, 1);

  if IOResult <> 0 then result := CotteErrModNotOpened
  else begin
    fsize := FileSize(f);
    ClausesCount := fsize div CotteModClauseSize - 2;

    if ((fsize mod CotteModClauseSize)<>0) or (ClausesCount<=0) then
      result := CotteErrModBadFormat
    else if ClausesCount > CotteModMaxClauses then result := CotteErrModTooLarge
    else if IOResult <> 0 then result := CotteErrModReadFault;

    if result = CotteOK then begin
      try SetLength(data, ClausesCount+2);
      except result := CotteErrNotEnoughRAM;
      end;

      if result = CotteOK then begin
        BlockRead(f, data[0], fsize, count);
        if (IOResult<>0) or (count<>fsize) then result := CotteErrModReadFault;
      end;
    end;
  end;

  close(f);

  if Result <> CotteOK then SetLength(data, 0)
  else begin
    swap := RawMode;
    RawMode := data;
    result := reload;

    if result <> CotteOK then begin
      RawMode := swap;
      finalize(data);
    end;
  end;
end;

//-----------------------------------------------------------------------------
function TotteMode.reload: integer;

  //. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .
  function chop(var s: ansistring): ansistring;
  begin
    result := '';

    while (s<>'') and (s[1]<>#09) do begin
      result := result + s[1];
      delete(s, 1, 1);
    end;

    if s<>'' then delete(s, 1, 1);
  end;
  //. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .
  function DecodeHeader(
    const data: TotteEncodedModClause;
    var Description: TotteModDesc;
    var CaseSensitive: boolean
  ): integer;

  var
    i: integer;
    s, desc, csflag: ansistring;

  begin
    result := CotteOK;

    s := '';
    for i := 0 to CotteModClauseSize-1 do s := s + char(data[i]);

    desc := chop(s);
    if length(desc) > CotteModDescLen then begin
      result := CotteErrModBadFormat;
      exit;
    end;

    csflag := chop(s);
    if (csflag<>'0') and (csflag<>'1') then begin
      result := CotteErrModBadFormat;
      exit;
    end;

    Description := desc;
    if csflag = '0' then CaseSensitive := false
    else if csflag = '1' then CaseSensitive := true
  end;
  //. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .
  function DecodeClause(
    const data: TotteEncodedModClause;
    CaseSensitive: boolean;
    var clause:  TotteModClause
  ): integer;

  var
    i: integer;
    s, rommod, tengmod, nextr, lastt, setlastt: ansistring;

  begin
    result := CotteOK;

    s := '';
    for i := 0 to CotteModClauseSize-1 do s := s + char(data[i]);    

    rommod := chop(s);
    if (rommod='') or (length(rommod)>CotteModStringLen) then begin
      result := CotteErrModBadFormat;
      exit;
    end;
   
    tengmod := chop(s);
    if (length(tengmod)>CotteModStringLen) then begin
      result := CotteErrModBadFormat;
      exit;
    end;
    
    nextr := chop(s);
    if (nextr='') then nextr := CotteAnyChar;

    lastt := chop(s);
    if (lastt='') then lastt := CotteAnyChar;
    
    setlastt := chop(s);
    if (setlastt='') then setlastt := CotteAnyChar;

    with clause do begin
      Roman := rommod;
      Tengwar := tengmod;
      NextRoman := nextr[1];
      LastTengwa := lastt[1];
      SetLastTengwa := setlastt[1];
      
      if CaseSensitive then begin
        Roman := self.AdjustStr(CotteRoman, Roman);
	NextRoman := self.AdjustChar[CotteRoman, NextRoman];
      end;
    end;
  end;
  //. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

var
  TempClauses: TotteModClauses;
  desc: TotteModDesc;
  csflag: boolean;
  i: integer;

begin
  if length(self.RawMode) < 3 then result := CotteErrModNotLoaded
  else begin
    try
      SetLength(TempClauses, length(self.RawMode)-2);
    except
      result := CotteErrNotEnoughRAM;
      SetLength(TempClauses, 0);
      exit;
    end;

    result := DecodeHeader(self.RawMode[0], desc, csflag);

    if result<>0 then begin
      SetLength(TempClauses, 0);
      exit;
    end;

    for i := 2 to High(self.RawMode) do begin
      result := DecodeClause(self.RawMode[i], csflag, TempClauses[i-2]);

      if result<>0 then begin
        SetLength(TempClauses, 0);
        exit;
      end;
    end;

    if result = CotteOK then begin
      self.Description := desc;
      self.CaseSensitive := csflag;
      self.clauses := TempClauses;
      self.AutoBackClauses;
      {
      Don't finalize TempClauses!
      That will erase our new self.clauses!
      For similar, don't finalize old self.clauses!
      }
    end
    else SetLength(TempClauses, 0);
  end;
end;

//-----------------------------------------------------------------------------
function SortBackClauses(const a, b: TotteModClause): boolean; register;

  function cl(c: char): integer;
  begin
    if c = CotteAnyChar then result := 2
    else if c = CotteNoAlpha then result := 1
    else result := CotteOK;
  end;

var
  c1tag, c2tag: integer;

begin
  if a.Roman[1] = b.Roman[1] then begin
    if a.Roman[0] = b.Roman[0] then begin
      if a.Roman = b.Roman then begin
        c1tag := 3*cl(a.NextRoman)+9*cl(a.LastTengwa)+cl(a.SetLastTengwa);
	c2tag := 3*cl(b.NextRoman)+9*cl(b.LastTengwa)+cl(b.SetLastTengwa);
	if c1tag = c2tag then result := a.Tengwar[0] > b.Tengwar[0]
	else result := c1tag > c2tag;
      end
      else result := a.Roman > b.Roman;
    end
    else result := a.Roman[0] < b.Roman[0];
  end
  else result := a.Roman[1] > b.Roman[1];
end;

//-----------------------------------------------------------------------------
// This procedure attempts to guess how to transcribe the tengwar text produced
// with the current mode file back to Roman. The result is stored in BackClauses,
// in the same way as normal mode file. Note: types of Clauses and BackClauses
// are identical, but fields of BackClauses have "inverted" meaning.
procedure TotteMode.AutoBackClauses;

var
  TempClauses: TotteModClauses;
  TempStr: TotteModString;
  FindByTengwa, FindByLetter: TotteCharAdjust;
  i, j: integer;

begin
  //Copy all significant records to a temporary location
  setlength(TempClauses, length(self.clauses));
  j := 0;

  for i := 0 to high(self.clauses) do
  if self.clauses[i].Tengwar <> '' then begin
    TempClauses[j] := self.clauses[i];

    if not self.TengwaCaseSensitive then
    with TempClauses[j] do begin
      Tengwar := self.AdjustStr(CotteTengwar, Tengwar);
      LastTengwa := self.AdjustChar[CotteTengwar, LastTengwa];
      SetLastTengwa := self.AdjustChar[CotteTengwar, SetLastTengwa];
    end;

    inc(j);
  end;

  setlength(TempClauses, j);

  //Establish a Tengwa-Roman relationship
  fillchar(FindByTengwa, sizeof(FindByTengwa), byte(CotteAnyChar));

  for i := 0 to high(TempClauses) do with TempClauses[i] do
    if (FindByTengwa[Tengwar[1]] = CotteAnyChar) and
       (length(Roman) = 1) and (length(Tengwar) = 1) then
      FindByTengwa[Tengwar[1]] := Roman[1];

  for i := 1 to 32 do FindByTengwa[char(i)] := char(i);

  //Now establish a Roman-Tengwa relationship
  fillchar(FindByLetter, sizeof(FindByLetter), byte(CotteAnyChar));

  for i := 0 to high(Clauses) do with Clauses[i] do
    if (FindByLetter[Roman[1]] = CotteAnyChar) and
       (length(Roman) = 1) and (length(Tengwar) = 1) then
      FindByLetter[Roman[1]] := Tengwar[1];

  for i := 1 to 32 do FindByLetter[char(i)] := char(i);

  //Now it's time to translate the records
  for i := 0 to High(TempClauses) do with TempClauses[i] do begin
    TempStr := Roman;
    Roman := Tengwar;
    Tengwar := TempStr;
    NextRoman := FindByLetter[NextRoman];
    LastTengwa := FindByTengwa[LastTengwa];
    SetLastTengwa := FindByTengwa[SetLastTengwa];
  end;

  self.SortClauses(TempClauses, @SortBackClauses);
  self.BackClauses := TempClauses;
end;

//===TotteTranscriber==========================================================
procedure TotteTranscriber.reset;
begin
  inherited;
  DoTranscription := true;
  RevNum := false;
  CirNum := false;
  DotNum := false;
  BarNum := false;
  B12Num := false;
  SepNum := true;
end;

//-----------------------------------------------------------------------------
constructor TotteTranscriber.create;
begin
  inherited;
  reset;
end;

//-----------------------------------------------------------------------------
function SortControl(const a, b: TotteModClause): boolean; register;
begin
  result := a.Roman[1] > b.Roman[1];
end;

//-----------------------------------------------------------------------------
function TotteTranscriber.GetReady: integer;
{
  Run this proc whenever you have just set all the params needed to do
  the transcription and before you start. If you don't, you'll get
  erroneous results!

  This proc prepares some data for transcription process. It also sorts
  and indexes the mode clauses by their first leotters to speed the
  transcription process up (about 1.5 times).
}
  //. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .
  function GetHalfReady(var Clauses: TotteModClauses;
                        var FirstInd, LastInd: TotteCharIndex): integer;
  var
    c: char;
    i: integer;

  begin
    if length(clauses) = 0 then begin
      result := CotteErrModNotLoaded;
      exit;
    end
    else result := CotteOK;

    //Sort the mode clauses
    self.SortClauses(clauses, SortControl);

    //Chars that do not open any mode rule are given strange
    //indexes to be immediately skipped during thanscription
    for c := #00 to #255 do FirstInd[c] := high(clauses);
    fillchar(LastInd, sizeof(LastInd), 0);

    //Index the mode clauses by the first char of the ROMAN.
    for i := 0 to high(clauses) do
      LastInd[clauses[i].Roman[1]] := i;
      
    for i := high(clauses) downto 0 do
      FirstInd[clauses[i].Roman[1]] := i;
  end;
  //. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

begin
  //Set characters skipped in transcription process
  fillchar(IsWhiteSpace, SizeOf(IsWhiteSpace), false);
  fillchar(IsWhiteSpace, 32, true);

  if SepNum then begin
    IsWhiteSpace[#32] := true;
    IsWhiteSpace[#160] := true;
  end;

  result := GetHalfReady(self.clauses, FirstInd, LastInd);
  if result = CotteOK then
    result := GetHalfReady(self.BackClauses, FirstBackInd, LastBackInd);
end;

//-----------------------------------------------------------------------------
function TotteTranscriber.LoadFromFile(const fname: ansistring): integer;
begin
  result := inherited LoadFromFile(fname);
  if result = CotteOK then result := GetReady;
end;

//-----------------------------------------------------------------------------
function TotteTranscriber.TranscribeLine(TranscribeBack: boolean; const s: ansistring): ansistring;

var
  buffer: ansistring;
  LastWrittenTengwa: char;
  index: integer;

  {$IFDEF PREALOCATE}
    ousindex: integer;
  {$ENDIF}

  //Either points to mode.clauses or mode.BackClauses
  clauses: TotteModClauses;

  //Either pointers to FirstRomanInd/LastRomanInd
  //or to FirstTengwarInd/LastTengwarInd
  FirstIndex, LastIndex: ^TotteCharIndex;

  //Either RomanQuestionMark or TengwarQuestionMark
  QuestionMark: char;

  //Indexes to IsAlpha[boolean].
  //Swapped when transcribing back: IsAlpha[AsRoman]
  //returns IsTengwar then, and thus these indexes
  //are separated from global otte.CotteRoman
  //and otte.CotteTengwar.
  AsRoman, AsTengwar: boolean;

  //. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .
  {$IFDEF PREALOCATE}
  { This is a workaround for slow string appending. It explicitly preallocates
    space for output string and then calls "move". It's mainly for Delphi, as
    Free Pascal with cmem under Linux has faster memory management. }
  procedure StrAdd(const ins: ansistring; var ousindex: Integer; var ous: ansistring);

  var
    count: integer;

  begin
    count := length(ins);
    if count <> 0 then begin
      if (ousindex+count > length(ous)) then SetLength(ous, ousindex+count+80);
      move(ins[1], ous[ousindex], count);
      inc(ousindex, count);
    end;
  end;
  {$ENDIF}
  //. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .
  function CutNumeral(const buffer: ansistring; var index: integer): ansistring;

  var
    charcount, numcount, i, j: integer;
    IsDigit, IsNonSpacing: ^TotteCharProperty;

  begin
    result := '';
    charcount := 0;
    numcount := 0;
    IsDigit := @self.IsDigit[AsRoman];
    IsNonSpacing := @self.IsNonSpacing[AsRoman];

    while (index+charcount <= length(buffer)) do begin
      if IsDigit[buffer[index+charcount]] then begin
        inc(charcount);
	inc(numcount);
      end
      else if IsNonSpacing[buffer[index+charcount]] then inc(charcount)
      else break;
    end;

    if charcount = numcount then result := copy(buffer, index, charcount)
    else begin
      setlength(result, numcount);
      j := 1;

      for i := index to index + charcount - 1 do
      if IsDigit[buffer[i]] then begin
        result[j] := buffer[i];
	inc(j);
      end;
    end;

    inc(index, charcount);
  end;
  //. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .
  procedure ReverseNumeral(var s: ansistring);

  var
    i, j, len: integer;
    c: char;

  begin
    len := length(s);
    j := len;

    for i := 1 to len div 2 do begin
      c := s[i];
      s[i] := s[j];
      s[j] := c;
      dec(j);
    end;
  end;
  //. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .
  function ProcessNumerals(const buffer: ansistring; var index: integer): ansistring;

  const
    //indexes for Nums[digit, index] array
    //Note for barover and barunder: they are taken in account only for first
    //written digit, as the bar may be longer or shorter. Then, for the second
    //and following digits, the standard, long bars are used.
    dotover  = 0;
    barover  = 1;
    dotunder = 2;
    barunder = 3;
    circle   = 4;

    // standard bars
    LongBarOver  = #236;  //over second and further digits
    LongBarUnder = #237;  //under second and further digits

    // Array of Elvish numerals and their "attributes"
    Nums: array[1..12, dotover..circle] of char =

     // dotover barover dotunder barunder circle      digit
        (('T',   #238,   #201,    #239,   #168),     {  0  }
         ('G',   #238,   #202,    #239,   #168),     {  1  }
         ('%',   #236,   #200,    #237,   #153),     {  2  }
         ('%',   #236,   #200,    #237,   #152),     {  3  }
         ('G',   #238,   #202,    #239,   #153),     {  4  }
         ('T',   #238,   #201,    #239,   #152),     {  5  }
         ('%',   #236,   #200,    #237,   #152),     {  6  }
         ('T',   #238,   #201,    #239,   #153),     {  7  }
         ('G',   #238,   #202,    #239,   #153),     {  8  }
         ('%',   #236,   #200,    #237,   #152),     {  9  }
         ('B',   #238,   #202,    #239,   #153),     {  10 }
         ('B',   #238,   #202,    #239,   #153));    {  11 }

  var
    NumBuff: ansistring;    //ASCII string that holds the number
    NumLen: integer;        //initial length of output numeral
    OutLen: integer;        //final length of output numeral
    i, j: integer;
    k: byte;                //NumBuff[i] as an index of Nums array
    CirclePos: integer;     //position of duodecades mark;

  begin
    NumBuff := CutNumeral(buffer, index);
    Result := '';

    //if transcribing back, then revert digit order BEFORE processing
    if TranscribeBack and RevNum then ReverseNumeral(NumBuff);

    //translate digits to #1 based: '0', '1' .. '9' => #1, #1 .. #10
    for i := 1 to length(NumBuff) do
      NumBuff[i] := char(DigitValues[AsRoman, NumBuff[i]]);

    //convert numeric base, if requested.
    if B12Num then
      NumBuff := OtteBaseConv(NumBuff, NumBase[AsRoman], NumBase[AsTengwar]);

    NumLen := Length(NumBuff);

    //translate digits from #1 based (#1, #2 ...) to target alphabet
    for i := 1 to NumLen do
      NumBuff[i] := Digits[AsTengwar, byte(NumBuff[i])];

    //write the numeral
    if TranscribeBack then Result := NumBuff
    else begin
      //do you want reverse digit order?
      if RevNum then begin
        ReverseNumeral(NumBuff);
	CirclePos := 2;
      end
      else CirclePos := NumLen - 1;

      //If no gadget to add, pass NumBuff through.
      //If only circle (duodecades mark) to add, simply insert it to NumBuff.
      //Otherwise make a buffer and build the output string in it.
      if not (DotNum or BarNum) then begin
        if CirNum and (CirclePos > 0) then
          insert(Nums[DigitValues[CotteTengwar, NumBuff[CirclePos]], circle],
	         NumBuff, CirclePos+1);

        result := NumBuff;
      end
      else begin
        //calculate the output buffer length
        OutLen := 1;
        if BarNum then inc(OutLen);
        if DotNum then inc(OutLen);
        OutLen := OutLen * NumLen;
        //Inc(OutLen);
        if CirNum and not (b12num and DotNum) then inc(OutLen);
        
        //try to allocate the output buffer
        setlength(result, OutLen);
        
        //construct the output numeral
        j := 1;
        for i := 1 to NumLen do begin

          //digit value
          k := DigitValues[CotteTengwar, NumBuff[i]];

          //add the digit
          Result[j] := NumBuff[i];
          inc(j);

          //if dots requested
          if DotNum then begin
            if B12Num then begin
              if (not CirNum) or (i <> CirclePos) then begin
                Result[j] := Nums[k, dotunder];
                inc(j);
              end;
            end
            else begin
              Result[j] := Nums[k, dotover];
              inc(j);
            end;
          end;

          //if bars requested
          if BarNum then begin
            if (i=1) then begin //first written digit - use dedicated bars
              if B12Num then Result[j] := Nums[k, barunder]
              else Result[j] := Nums[k, barover];
            end
            else begin //use standard long bars
              if B12Num then Result[j] := LongBarUnder
              else Result[j] := LongBarOver;
            end;
            inc(j);
          end;

          //if  circled duodecades requested
          if CirNum and (i = CirclePos) then begin
            Result[j] := Nums[k, circle];
            inc(j);
          end;
        end;

        setlength(result, j-1);
      end;
    end;
  end;
  //. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .
  function CharsNotMatch(ch, ch_mod: char; AsRoman: boolean): boolean; register;
  begin
    Result := (ch_mod <> CotteAnyChar) and (ch <> ch_mod) and
              ((ch_mod <> CotteNoAlpha) or IsAlpha[AsRoman, ch]);
  end;
  //. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .
  function StrStart(s1, s2: PChar; n: integer): boolean;
  begin
    result := false;
    while n <> 0 do begin
      if s1^ <> s2^ then exit;
      inc(s1);
      inc(s2);
      dec(n);
    end;
    result := true;
  end;
  //. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .
  function ProcessText(const buffer: ansistring; var index: integer): TotteModString;

  var
    i, Ilen, Rlen: Integer;
    modclause: ^TotteModClause;
    Temp: PChar;

  begin
    Result := '';
    Temp := @buffer[index];
    Ilen := length(buffer) - index + 1;
    modclause := addr(clauses[FirstIndex[Temp^]]);
    Dec(modclause); //don't dereference this now, may be out of range!

    //here GetReady indexing efforts take effect
    //we don't have to search through all the clauses
    for i := FirstIndex[Temp^] to LastIndex[Temp^] do begin
      Inc(modclause); //now it is surely in the range

      with modclause^ do begin
        //further speed improvement - great for complicated modes (i.e. Quenya)
	//caution - checking characters no. 4 and next slows things down...
	//takes no effect for simple modes (i. e. Sindarin_beleriand)
	Rlen := length(Roman);

	if (Rlen > Ilen) or
	   ((Rlen >= 2) and (Roman[2] <> Temp[1])) or
           ((Rlen >= 3) and (Roman[3] <> Temp[2])) or
	   CharsNotMatch(LastWrittenTengwa, LastTengwa, AsTengwar) or
	   CharsNotMatch(Temp[Length(Roman)], NextRoman, AsRoman) then continue
        else
	if (Rlen <= 3) or StrStart(@Roman[4], @Temp[3], length(roman)-3) then
        begin
          result := Tengwar;
          LastWrittenTengwa := SetLastTengwa;
	  inc(index, length(Roman));
          exit;
        end;
      end;
    end;

    //when transcription error...
    if QuestionMark <> CotteNoQuestionMark then begin
      result := QuestionMark;
      LastWrittenTengwa := ' ';
    end;

    inc(index);
  end;
  //. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

begin
  AsRoman := not TranscribeBack;
  AsTengwar := not AsRoman;

  if AsRoman then begin
    QuestionMark := TengwarQuestionMark;
    clauses := self.Clauses;
    FirstIndex := @FirstInd;
    LastIndex := @LastInd;
    
    if CaseSensitive then buffer := s
    else buffer := AdjustStr(CotteRoman, s);
  end
  else begin
    QuestionMark := RomanQuestionMark;
    clauses := BackClauses;
    FirstIndex := @FirstBackInd;
    LastIndex := @LastBackInd;

    if TengwaCaseSensitive then buffer := s
    else buffer := AdjustStr(CotteTengwar, s);
  end;

  result := '';
  if length(clauses) = 0 then exit;

  index := 1;
  
  {$IFDEF PREALOCATE}
    ousindex := 1;
    setLength(result, length(s));
  {$ENDIF}
  
  LastWrittenTengwa := ' ';

  //if transcription was cancelled before
  if not DoTranscription then DoTranscription := true;

  while (index <= length(buffer)) and DoTranscription do begin
    if IsWhiteSpace[buffer[index]] then begin
      {$IFDEF PREALOCATE}StrAdd(buffer[index], ousindex, result);
      {$ELSE}result := result + buffer[index];
      {$ENDIF}
      inc(index);
      LastWrittenTengwa := ' ';
    end
    else if SepNum and IsDigit[AsRoman, buffer[index]] then begin
      {$IFDEF PREALOCATE}StrAdd(ProcessNumerals(buffer, index), ousindex, result);
      {$ELSE}result := result + ProcessNumerals(buffer, index);
      {$ENDIF}
      LastWrittenTengwa := CotteNoAlpha;
    end
    else begin
      {$IFDEF PREALOCATE}StrAdd(ProcessText(buffer, index), ousindex, result);
      {$ELSE}result := result + ProcessText(buffer, index);
      {$ENDIF}
    end;
  end;

  {$IFDEF PREALOCATE}SetLength(result, ousindex-1);{$ENDIF}
  
  //if transcription was cancelled in progress
  if not DoTranscription then begin
    DoTranscription := true;
    result := s;
  end;
end;

//-----------------------------------------------------------------------------
procedure TotteTranscriber.StopTranscription;
begin
  DoTranscription := false;
end;

//===EndOfImplementation=======================================================
end.
