{-----------------------------------------------------------------------------

  Unit otte_math manages conversion between numeric systems. The code
  cannot convert numbers longer than 19-20 decimal (or 16-17 duodecimal)
  digits, as converted number must fit within an 64-bit unsigned integer.

                                --*--

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

------------------------------------------------------------------------------}

{$I otte_cfg.pas}
unit otte_math;

interface //===================================================================

const
  CotteMinNumBase = 2;
  CotteMaxNumBase = 36;

type TotteNumBase = CotteMinNumBase .. CotteMaxNumBase;

function OtteBaseConv(const num: ansistring; inbase, outbase: TotteNumBase): ansistring;

implementation //==============================================================

{ 'Powers' table keeps digits weight for all supported numeric bases.
  First index denotes numeric base, second index denotes digit position
  in a numeric string. For example: Powers[10, 3] means 100, as the third
  less meaning digit of decimal based numeric string has weight 100.
  
  Powers[10, 64], though, means 0, because OTTE doesn't support such long
  decimal strings. MaxNum table stores the maximal QWord value represented
  in all supported numeric bases.
}
const
  MaxInQWord = 63; //max length of the input numeral, longer will be cut.

type
  QWord = Int64; //Damn! Delphi doesn't support QWords!
  TRetString = string[MaxInQWord];

var
  MaxNum: array[TotteNumBase] of ansistring;
  powers: array[TotteNumBase, 1..63] of QWord;

//-----------------------------------------------------------------------------
function QWordToStr(a: QWord; base: TotteNumBase): TRetString;

var
  i: integer;

begin
  fillchar(result[1], MaxInQword, #01);
  result[0] := char(MaxInQword);
  i := MaxInQword;

  while a <> 0 do begin
    result[i] := char(1 + byte(a mod byte(base)));
    a := a div byte(base);
    dec(i);
  end;

  //remove leading zeroes
  delete(result, 1, i);
end;

//-----------------------------------------------------------------------------
function OtteBaseConv(const num: ansistring; InBase, OutBase: TotteNumBase): ansistring;

var
  i, count: integer;
  zeroes: ansistring; //keeps leading zeroes if any
  MaxStrLen: byte;
  NumVal: QWord;

begin
  result := num;

  //move leading zeroes into safe place, if there are any
  count := 0;
  while (count < length(result)) and (result[count+1] = #01) do inc(count);
  if count > 0 then begin
    zeroes := Copy(result, 1, count);
    Delete(result, 1, count);
  end;

  if length(result) > 0 then
  begin
    //isn't the numeral too big?
    MaxStrLen := length(MaxNum[inbase]);
    if length(result) > MaxStrLen then begin
      setlength(result, MaxStrLen);
      if result > MaxNum[inbase] then setlength(result, length(result)-1);
    end;

    //do the conversion
    NumVal := 0;

    for i := 1 to length(result) do
      NumVal := NumVal + powers[inbase, length(result)+1-i] * (byte(result[i])-1);

    result := QwordToStr(NumVal, OutBase);
  end;

  { Restore leading zeroes, if there were any before the conversion.
    NOTE: This is of course an error in fractions :(
    I want to add "IsFraction" parameter in future versions,
    but I need Galadhorn's advice before this...
  }
  result := zeroes + result;
end;

//-----------------------------------------------------------------------------
var
  base, i: byte;
  basepow: QWord;

begin
  {INITIALIZATION OF MAXNUM TABLE}
  for i := CotteMinNumBase to CotteMaxNumBase do
    MaxNum[i] := QWordToStr(high(Qword), i);

  {INITIALIZATION OF POWERS TABLE}
  fillchar(powers, sizeof(powers), 0);

  for base := CotteMinNumBase to CotteMaxNumBase do begin
    basepow := 1;

    for i := 1 to length(MaxNum[base]) do begin
      powers[base, i] := basepow;
      if i <> length(MaxNum[base]) then basepow := basepow * base;
    end;
  end;
end.
